from pathlib import Path

import tensorrt as trt

from . import dynamic_axis


class ONNX2TensorRT:

    def __init__(self, onnx_path, dynamic_axes: list[dynamic_axis.DynamicAxisInfo] = None):
        self._onnx_path = onnx_path
        self._dynamic_axes = dynamic_axes

    def convert(self, engine_path):

        def is_dim_dynamic(dim):
            is_dim_str = not isinstance(dim, int)
            return dim is None or is_dim_str or dim < 0

        def is_dynamic_tensor(tensor: trt.ITensor) -> bool:
            return len([dim for dim in tensor.shape if is_dim_dynamic(dim)]) > 0

        logger = trt.Logger(trt.Logger.INFO)
        builder = trt.Builder(logger)
        network: trt.INetworkDefinition = builder.create_network(
            1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH))
        parser = trt.OnnxParser(network, logger)
        parse_result = parser.parse_from_file(self._onnx_path)

        num_outputs = network.num_outputs
        for idx in range(num_outputs):
            tensor = network.get_output(idx)
            print(f"output:{tensor.name}")

        if not parse_result:
            logger.log(trt.Logger.ERROR, "onnx parse error")
            for err_idx in parser.num_errors:
                logger.log(trt.Logger.ERROR, parser.get_error(err_idx))
            raise RuntimeError("Failed to parse ONNX model")

        config: trt.IBuilderConfig = builder.create_builder_config()
        if self._dynamic_axes:
            self._dynamic_axes = {item.input_name: item for item in self._dynamic_axes}
            logger.log(trt.Logger.INFO, "start to bind profile config")
            profile: trt.IOptimizationProfile = builder.create_optimization_profile()
            num_inputs = network.num_inputs
            for idx in range(num_inputs):
                tensor: trt.ITensor = network.get_input(idx)
                if tensor.is_shape_tensor:
                    if tensor.name in self._dynamic_axes:
                        dynamic_axis = self._dynamic_axes[tensor.name]
                        profile.set_shape_input(dynamic_axis.input_name, dynamic_axis.min_shape, dynamic_axis.opt_shape,
                                                dynamic_axis.max_shape)
                    else:
                        raise RuntimeError(f"{tensor.name} is shape tensor, must provide profile values")
                elif is_dynamic_tensor(tensor):
                    if tensor.name in self._dynamic_axes:
                        dynamic_axis = self._dynamic_axes[tensor.name]
                        profile.set_shape(dynamic_axis.input_name, dynamic_axis.min_shape, dynamic_axis.opt_shape,
                                          dynamic_axis.max_shape)
                    else:
                        raise RuntimeError(f"{tensor.name} is dynamic tensor, must provide profile shapes")

                else:
                    logger.log(trt.Logger.INFO, f"tensor: {tensor.name} is static tensor")
            result_code = config.add_optimization_profile(profile)
            if result_code == -1:
                raise RuntimeError("bind profile error, input is not valid, check that")
            logger.log(trt.Logger.INFO, "complete to bind profile config")

        serialized_engine = builder.build_serialized_network(network, config)
        if serialized_engine is None:
            raise RuntimeError("build network error")
        parent_dir = Path(engine_path).parent
        if not parent_dir.exists():
            parent_dir.mkdir(parents=True, exist_ok=True)
        with open(engine_path, 'wb') as f:
            f.write(serialized_engine)
        logger.log(trt.Logger.INFO, "complete build trt engine")



class DynamicAxisInfo:
    def __init__(self, name: str, min_value: tuple, opt_value: tuple, max_value: tuple):
        """
        封装动态轴信息
        :param name: 输入张量的名称
        :param min_value: 最小输入形状
        :param opt_value: 优化输入形状
        :param max_value: 最大输入形状
        """
        self.__name = name  # 私有属性，输入张量的名称
        self.__min = min_value  # 私有属性，动态输入的最小形状
        self.__opt = opt_value  # 私有属性，动态输入的优化形状
        self.__max = max_value  # 私有属性，动态输入的最大形状

    @property
    def input_name(self) -> str:
        """获取输入张量的名称"""
        return self.__name

    @property
    def min_shape(self) -> tuple:
        """获取动态输入的最小形状"""
        return self.__min

    @property
    def opt_shape(self) -> tuple:
        """获取动态输入的优化形状"""
        return self.__opt

    @property
    def max_shape(self) -> tuple:
        """获取动态输入的最大形状"""
        return self.__max



from typing import List

import numpy as np
import onnxruntime as onnxrt

from ...base import ABSBackend
from ...config import Device


class ONNXBackend(ABSBackend):

    def __init__(self, devices: List[Device], model_path: str):
        self._devices = devices
        self._model_path = model_path
        support_providers = onnxrt.get_available_providers()
        providers = []
        for device in devices:
            if device in support_providers:
                providers.append(device)
            else:
                for ele in support_providers:
                    if device.value.lower() in ele.lower() or ele.lower() in device.value.lower():
                        providers.append(ele)
        if not providers:
            providers.append("CPUExecutionProvider")
        options = onnxrt.SessionOptions()
        self._session = onnxrt.InferenceSession(model_path, providers=providers, sess_options=options)

    def run(self, inputs: dict[str, np.ndarray]) -> dict[str, np.ndarray]:
        inference_outputs = self._session.run(None, inputs)
        output_nodes = self._session.get_outputs()
        results = {}
        for node, out in zip(output_nodes, inference_outputs):
            results[node.name] = out
        return results

    def release(self):
        pass



from pathlib import Path

import numpy as np
import tensorrt as trt
from cuda import cudart

from .convert.dynamic_axis import DynamicAxisInfo
from .convert.onnx_tensorrt import ONNX2TensorRT
from .utils import cuda_call, get_io, \
    copy_data_to_gpu, free_device_memory, synchronize_stream
from ...base import ABSBackend


class OutputAllocator(trt.IOutputAllocator):

    def __init__(self):
        trt.IOutputAllocator.__init__(self)
        self.buffers = {}
        self.shapes = {}

    def reallocate_output(self, tensor_name, memory, size, alignment):
        if tensor_name in self.buffers:
            del self.buffers[tensor_name]
        device_ptr = cuda_call(cudart.cudaMalloc(size))
        self.buffers[tensor_name] = device_ptr
        return int(device_ptr)

    def notify_shape(self, tensor_name, shape):
        self.shapes[tensor_name] = tuple(shape)


class TensorRTBackend(ABSBackend):
    """
    TensorRT推理后端
    """

    def __init__(self, model_path, engine_path, dynamic_axes: list[DynamicAxisInfo] = None):
        self._model_path = model_path
        self._engine_path = engine_path
        if not Path(self._engine_path).exists():
            ONNX2TensorRT(self._model_path, dynamic_axes).convert(engine_path=self._engine_path)

        self._logger = trt.Logger(trt.Logger.INFO)
        self._runtime = trt.Runtime(self._logger)
        with open(self._engine_path, 'rb') as f:
            self._engine: trt.ICudaEngine = self._runtime.deserialize_cuda_engine(f.read())
        self._context: trt.IExecutionContext = self._engine.create_execution_context()

    def run(self, inputs: dict[str, np.ndarray]) -> dict[str, np.ndarray]:
        stream = cuda_call(cudart.cudaStreamCreate())

        input_buffers = {}
        for tensor_name in get_io(self._engine, trt.TensorIOMode.INPUT):
            array = inputs[tensor_name]
            if self._engine.is_shape_inference_io(tensor_name):
                ptr = array.ctypes.data
            else:
                host_dtype = array.dtype
                tensor_dtype = self._engine.get_tensor_dtype(tensor_name)
                if trt.nptype(tensor_dtype) != host_dtype:
                    raise RuntimeError(
                        f"input buffer {tensor_name} type {host_dtype} not match model tensor typpe {trt.nptype(tensor_dtype)}")
                ptr = copy_data_to_gpu(array, stream)
                if tensor_name in input_buffers:
                    free_device_memory(input_buffers[tensor_name])
                input_buffers[tensor_name] = ptr
            if self._context.get_tensor_address(tensor_name) != ptr:
                self._context.set_tensor_address(tensor_name, ptr)
            shape = array.shape
            if self._context.get_tensor_shape(tensor_name) != shape:
                self._context.set_input_shape(tensor_name, shape)

        output_allocator = OutputAllocator()
        for tensor_name in get_io(self._engine, trt.TensorIOMode.OUTPUT):
            self._context.set_output_allocator(tensor_name, output_allocator)

        # 执行运算
        self._context.execute_async_v3(stream_handle=stream)

        # 从CUDA复制数据到主存
        outputs = {}
        for tensor_name in get_io(self._engine, trt.TensorIOMode.OUTPUT):
            dtype = trt.nptype(self._engine.get_tensor_dtype(tensor_name))
            shape = output_allocator.shapes[tensor_name]
            array: np.ndarray = np.empty(shape, dtype)
            array_ptr = array.ctypes.data
            ptr = output_allocator.buffers[tensor_name]
            cuda_call(cudart.cudaMemcpyAsync(array_ptr, ptr, array.nbytes,
                                             cudart.cudaMemcpyKind.cudaMemcpyDeviceToHost, stream))
            outputs[tensor_name] = array
        synchronize_stream(stream)
        for _, devPtr in input_buffers.items():
            free_device_memory(devPtr)
        for _, devPtr in output_allocator.buffers.items():
            free_device_memory(devPtr)
        return outputs

    def release(self):
        del self._engine
        del self._context



from abc import ABC, abstractmethod

import numpy as np


class ABSBackend(ABC):

    @abstractmethod
    def run(self, inputs: dict[str, np.ndarray]) -> dict[str, np.ndarray]:
        """
        执行推理
        :param inputs: key:输入名称 value:输入数据
        :return:       key:输出名称 value:输出数据
        """
        pass

    @abstractmethod
    def release(self):
        pass



import numpy as np
from .config import *


class Runtime:

    def __init__(self, config: Config):
        self._config: Config = config
        if not self._config.model_path:
            raise RuntimeError("model path cannot empty")
        if config.backend == Backend.TensorRT:
            from .backend.trt.tensorrt_backend import TensorRTBackend
            self._backend = TensorRTBackend(config.model_path, config.engine_path, config.dynamic_axes)
        else:
            from .backend.onx.onnx_backend import ONNXBackend
            self._backend = ONNXBackend(config.devices, config.model_path)

    def run(self, inputs: dict[str, np.ndarray]) -> dict[str, np.ndarray]:
        return self._backend.run(inputs)

    def release(self):
        self._backend.release()



from setuptools import setup, find_packages

setup(name="inferx",
      version="1.0.5",
      author="yuanyang",
      author_email="yuan823978@gmail.com",
      url="https://github.com/yuanyang1991/InferX",
      packages=find_packages(),
      description="deep learning deploy framework",
      python_requires='>=3.6',
      install_requires=[
          'tensorrt==10.3.0',
      ]
      )



from enum import Enum
from typing import List

from .backend.trt.convert.dynamic_axis import DynamicAxisInfo


class Backend(Enum):
    TensorRT = 1
    Onnx = 2


class Device(Enum):
    CPU = "cpu"
    GPU = "cuda"


class Config:

    def __init__(self, model_path: str,
                 engine_path: str = None,
                 backend: Backend = Backend.TensorRT,
                 devices: List[Device] = None,
                 dynamic_axes: list[DynamicAxisInfo] = None
                 ):
        """
        初始化框架的相关配置
        :param model_path:  onnx模型地址
        :param engine_path: tensorrt engine地址。如果使用ONNX后端推理，此参数可不传。
        :param devices:      当使用ONNX后端时，指定使用CPU或者CUDA推理
        :param backend:      选取使用的推理后端
        :param dynamic_axes: 当选取TensorRT推理后端且需要模型转换时，用于配置模型的动态轴信息
        """

        self._model_path = model_path
        self._engine_path = engine_path
        self._devices = devices
        self._backend = backend
        self._dynamic_axes = dynamic_axes

    @property
    def model_path(self) -> str:
        return self._model_path

    @property
    def engine_path(self) -> str:
        return self._engine_path

    @property
    def devices(self) -> List[Device]:
        return self._devices

    @property
    def backend(self) -> Backend:
        return self._backend

    @property
    def dynamic_axes(self) -> list[DynamicAxisInfo]:
        return self._dynamic_axes



import numpy as np
import tensorrt as trt
from cuda import cuda, cudart


def check_cuda_err(err):
    if isinstance(err, cuda.CUresult):
        if err != cuda.CUresult.CUDA_SUCCESS:
            raise RuntimeError("Cuda Error: {}".format(err))
    if isinstance(err, cudart.cudaError_t):
        if err != cudart.cudaError_t.cudaSuccess:
            raise RuntimeError("Cuda Runtime Error: {}".format(err))
    else:
        raise RuntimeError("Unknown error type: {}".format(err))


def cuda_call(call):
    err, res = call[0], call[1:]
    check_cuda_err(err)
    if len(res) == 1:
        res = res[0]
    return res


def get_io(engine: trt.ICudaEngine, mode):
    nums = engine.num_io_tensors
    result = []
    for idx in range(nums):
        tensor_name = engine.get_tensor_name(idx)
        if engine.get_tensor_mode(tensor_name) == mode:
            result.append(tensor_name)
    return result


def copy_data_to_gpu(array: np.ndarray, stream):
    def is_contiguous(obj: np.ndarray):
        return obj.flags["C_CONTIGUOUS"]

    if not is_contiguous(array):
        array = np.ascontiguousarray(array)
    device_ptr = cuda_call(cudart.cudaMalloc(array.nbytes))
    host_ptr = array.ctypes.data
    cuda_call(cudart.cudaMemcpyAsync(device_ptr, host_ptr, array.nbytes,
                                     cudart.cudaMemcpyKind.cudaMemcpyHostToDevice, stream))

    return device_ptr


def free_device_memory(device_ptr):
    cuda_call(cudart.cudaFree(device_ptr))


def synchronize_stream(stream):
    cuda_call(cudart.cudaStreamSynchronize(stream))
